from .abstract_accelerator import Accelerator, AcceleratorType

try:
    import torch.cuda
except ImportError:
    pass


class DIPU_Accelerator(Accelerator):
    """Accelerator for CUDA device.

    Args:
        Accelerator (Accelerator): _description_
    """

    def __init__(self) -> None:
        self._name_str = "cuda"
        self._communication_backend_name = "nccl"
        self.amp = self.get_amp()
        self.memory = torch.cuda.memory
        self._find_or_mock_module("flash_attn_2_cuda")

    def _find_or_mock_module(self, module_name) -> bool:
        import importlib.util
        import sys
        import types

        """Find or mock a module. Return True if the module is found, False otherwise."""
        module_spec = importlib.util.find_spec(module_name)
        if module_spec is None:
            sys.modules[module_name] = types.SimpleNamespace()  # type: ignore
        return module_spec is not None

    def get_backend_name(self):
        """
        Return the name of the accelerator.
        """
        return self._name_str

    def get_accelerator_backend(self):
        """
        Return the name of the backend.
        """
        return AcceleratorType.DIPU

    # Device APIs
    def device_name(self, device_index=None):
        """
        Return the name of the device.
        """
        if device_index is None:
            return "cuda"
        return "cuda:{}".format(device_index)

    def set_device(self, device_index):
        """
        Bind the current process to a device.
        """
        torch.cuda.set_device(device_index)

    def get_device_id(self):
        """
        Return the current device index.
        """
        return torch.cuda.current_device()

    def current_device_name(self):
        """
        Return the name of the current device.
        """
        return "cuda:{}".format(torch.cuda.current_device())

    def device_count(self):
        """
        Return the number of devices on the machine.
        """
        return torch.cuda.device_count()

    def synchronize(self, device_index=None):
        """
        Synchronize the current process.
        """
        return torch.cuda.synchronize(device_index)

    # RNG APIs
    def random(self):
        """
        Get random number.
        """
        return torch.random

    def set_rng_state(self, new_state, device_index=None):
        """
        Sets the random number generator state of the specified GPU.
        """
        if device_index is None:
            return torch.cuda.set_rng_state(new_state)

        return torch.cuda.set_rng_state(new_state, device_index)

    def get_rng_state(self, device_index=None):
        """
        Returns the random number generator state of the specified GPU as a ByteTensor.
        """
        if device_index is None:
            return torch.cuda.get_rng_state()

        return torch.cuda.get_rng_state(device_index)

    def manual_seed(self, seed):
        """
        Sets the seed for generating random numbers for the current GPU.
        """
        return torch.cuda.manual_seed(seed)

    def manual_seed_all(self, seed):
        """
        Set the random seed for the all processes.
        """
        return torch.cuda.manual_seed_all(seed)

    def initial_seed(self):
        """
        Returns the current random seed of the current GPU.
        """
        return torch.cuda.initial_seed()

    def default_generator(self, device_index):
        """
        Returns the default generators according to device index
        """
        return torch.cuda.default_generators[device_index]

    # Streams/Events
    @property
    def Stream(self):
        """
        A CUDA stream is a linear sequence of execution that belongs to
        a specific device, independent from other streams.
        See cuda-semantics for details.
        """
        return torch.cuda.Stream

    def stream(self, _stream):
        """
        Wrapper around the Context-manager StreamContext that selects a given stream.
        """
        return torch.cuda.stream(_stream)

    def current_stream(self, device_index=None):
        """
        Returns the currently selected Stream for a given device.
        """
        return torch.cuda.current_stream(device_index)

    def default_stream(self, device_index=None):
        """
        Returns the default Stream for a given device.
        """
        return torch.cuda.default_stream(device_index)

    @property
    def Event(self):
        """
        CUDA events are synchronization markers that can be used
        to monitor the device's progress, to accurately measure timing,
        and to synchronize CUDA streams.
        """
        return torch.cuda.Event

    # Memory management
    def empty_cache(self):
        """
        Releases all unoccupied cached memory currently held by the caching allocator
        so that those can be used in other GPU application and visible in nvidia-smi.
        """
        return torch.cuda.empty_cache()

    def memory_allocated(self, device_index=None):
        """
        Returns the current GPU memory occupied by tensors in bytes for a given device.
        """
        return torch.cuda.memory_allocated(device_index)

    def max_memory_allocated(self, device_index=None):
        """
        Returns the maximum GPU memory occupied by tensors in bytes for a given device.
        """
        return torch.cuda.max_memory_allocated(device_index)

    def reset_max_memory_allocated(self, device_index=None):
        """
        Resets the starting point in tracking maximum GPU memory occupied by
        tensors for a given device.
        """
        return torch.cuda.reset_max_memory_allocated(device_index)

    def memory_cached(self, device_index=None):
        """
        Returns the cached memory
        """
        return torch.cuda.memory_cached(device_index)

    def max_memory_cached(self, device_index=None):
        """
        Returns the maximum cached memory
        """
        return torch.cuda.max_memory_cached(device_index)

    def reset_max_memory_cached(self, device_index=None):
        """
        Resets the starting point in tracking maximum GPU memory managed by
        the caching allocator for a given device.
        """
        return torch.cuda.reset_max_memory_cached(device_index)

    def memory_stats(self, device_index=None):
        """
        Returns the memory stats
        """
        if hasattr(torch.cuda, "memory_stats"):
            return torch.cuda.memory_stats(device_index)

    def reset_peak_memory_stats(self, device_index=None):
        """
        Resets the "peak" stats tracked by the CUDA memory allocator.
        """
        if hasattr(torch.cuda, "reset_peak_memory_stats"):
            return torch.cuda.reset_peak_memory_stats(device_index)

    def memory_reserved(self, device_index=None):
        """
        Returns the current GPU memory managed by the caching allocator
        in bytes for a given device.
        """
        if hasattr(torch.cuda, "memory_reserved"):
            return torch.cuda.memory_reserved(device_index)

    def max_memory_reserved(self, device_index=None):
        """
        Returns the maximum GPU memory managed by the caching allocator
        in bytes for a given device.
        """
        if hasattr(torch.cuda, "max_memory_reserved"):
            return torch.cuda.max_memory_reserved(device_index)

    def total_memory(self, device_index=None):
        """
        Returns the total memory
        """
        return torch.cuda.get_device_properties(device_index).total_memory

    # Data types
    def is_bf16_supported(self):
        """
        Returns true if bf16 is supported. Otherwise, returns false
        """
        return torch.cuda.is_bf16_supported()

    def is_fp16_supported(self):
        """
        Returns true if fp16 is supported. Otherwise, returns false
        """
        major, _ = torch.cuda.get_device_capability()
        return bool(major >= 7)

    # Misc
    def get_amp(self):
        """
        Returns the 'amp' module from torch.cuda if available, else returns None.
        """
        if hasattr(torch.cuda, "amp"):
            return torch.cuda.amp
        return None

    def is_available(self):
        """
        Checks and returns True if CUDA is available, False otherwise.
        """
        return torch.cuda.is_available()

    def range_push(self, msg):
        """
        If available, pushes a range with the given message for profiling using NVTX.
        """
        if hasattr(torch.cuda.nvtx, "range_push"):
            return torch.cuda.nvtx.range_push(msg)

    def range_pop(self):
        """
        If available, pops the most recent range pushed using NVTX.
        """
        if hasattr(torch.cuda.nvtx, "range_pop"):
            return torch.cuda.nvtx.range_pop()

    def lazy_call(self, callback):
        """
        Executes the given callback with lazy propagation if available.
        """
        return torch.cuda._lazy_call(callback)

    def communication_backend_name(self):
        """
        Returns the name of the current communication backend.
        """
        return self._communication_backend_name

    # Tensor operations

    @property
    def BFloat16Tensor(self):
        """
        Property to get the BFloat16Tensor class from torch.cuda.
        """
        return torch.cuda.BFloat16Tensor

    @property
    def ByteTensor(self):
        """
        Property to get the ByteTensor class from torch.cuda.
        """
        return torch.cuda.ByteTensor

    @property
    def DoubleTensor(self):
        """
        Property to get the DoubleTensor class from torch.cuda.
        """
        return torch.cuda.DoubleTensor

    @property
    def FloatTensor(self):
        """
        Property to get the FloatTensor class from torch.cuda.
        """
        return torch.cuda.FloatTensor

    @property
    def HalfTensor(self):
        """
        Property to get the HalfTensor class from torch.cuda.
        """
        return torch.cuda.HalfTensor

    @property
    def IntTensor(self):
        """
        Property to get the IntTensor class from torch.cuda.
        """
        return torch.cuda.IntTensor

    @property
    def LongTensor(self):
        """
        Property to get the LongTensor class from torch.cuda.
        """
        return torch.cuda.LongTensor

    def pin_memory(self, tensor):
        """
        Pins the memory of the given tensor, if it's a CUDA tensor.
        """
        return tensor.pin_memory()

    def on_accelerator(self, tensor):
        """
        Checks and returns True if the given tensor is on an accelerator (CUDA device), False otherwise.
        """
        device_str = str(tensor.device)
        return bool(device_str.startswith("cuda:"))

    def set_allow_tf32(self, enable: bool):
        """
        Sets the `allow_tf32` flag in cuDNN and CUDA matrix multiplication to the given boolean value.
        """
        print(f"Not support tf32 for DIPU, {enable}!")

    def return_custom_bwd(self):
        """
        Returns the custom backward hook function from torch.cuda.amp, if available.
        """
        return torch.cuda.amp.custom_bwd

    def return_custom_fwd(self):
        """
        Returns the custom forward hook function from torch.cuda.amp, if available.
        """
        return torch.cuda.amp.custom_fwd



from typing import Dict, Tuple

import torch

from internlm.core.context.parallel_context import ParallelMode
from internlm.core.context.parallel_context import global_context as gpc
from internlm.model.modules.utils import is_moe_param
from internlm.utils.parallel import is_tensor_data_parallel_parameter, is_using_isp


def split_params_into_different_groups_for_optimizer(
    param_groups: Tuple[Dict],
) -> Tuple[Dict]:
    """Split parameters into different groups for optimizer

    Args:
        param_groups (Tuple[Dict]): The list of parameter groups to split
        Input Example:
        >>> (
        >>>     {'name': 'default', 'params': [tensor], 'weight_decay' :xxx},
        >>> )

    Returns:
        Tuple[Dict]: list of params groups for optimizer
        Output Example:
        >>> (
        >>>     {'name': 'default', 'params': [tensor], 'weight_decay' :xxx},
        >>>     {'name': 'embed_head', 'params': [tensor], 'weight_decay' :xxx},
        >>>     {'name': 'fp32', 'params': [tensor], 'weight_decay' :xxx},
        >>> )
    """

    if isinstance(param_groups, tuple):
        param_groups = list(param_groups)  # Tuple cannot be modified
    elif isinstance(param_groups, dict):
        param_groups = [param_groups]
    elif not isinstance(param_groups, list):
        raise ValueError(f"Unknown param group type of {type(param_groups)}")

    # create new groups for IS_TENSOR_DATA_PARALLEL parameter group
    new_groups = {}
    if is_using_isp():
        new_groups["embed_head"] = {"name": "embed_head", "params": [], "optimizer_mode": ParallelMode.DATA}
    # create new groups for fp32 parameter group
    new_groups["fp32"] = {"name": "fp32", "params": [], "optimizer_mode": ParallelMode.ZERO1}

    if gpc.config.model.get("num_experts", 1) > 1:
        for key in gpc.expert_parallel_group_names:
            new_groups[key] = {"name": key, "moe": True, "params": [], "optimizer_mode": ParallelMode.EXPERT_DATA}

    for pgroup in param_groups:
        # copy attribute from origin group, we assume the input param_groups only
        # have one group, so the attribute will not be copyed multiple times.
        for ori_key in pgroup.keys():
            if ori_key not in ("name", "params"):
                for _, group in new_groups.items():
                    group[ori_key] = pgroup[ori_key]
        # assign param
        origin_params = []
        for param in pgroup["params"]:
            if is_tensor_data_parallel_parameter(param):
                # should not be here if not isp mode
                new_groups["embed_head"]["params"].append(param)
            # moe param means MoE is enabled
            elif is_moe_param(param):
                new_groups[param.group_name]["params"].append(param)
            elif param.dtype == torch.float32 and gpc.config.model.dtype != torch.float32:
                new_groups["fp32"]["params"].append(param)
            else:
                origin_params.append(param)

        # default param group, which is the first group in the param groups
        pgroup["params"] = origin_params
        pgroup["optimizer_mode"] = ParallelMode.ZERO1

    # param groups may contain empty groups, such as fp32
    param_groups.extend(new_groups.values())

    return tuple(param_groups)


def create_param_groups(model, weight_decay):
    parameters = {
        "params": [param for param in model.parameters() if param.requires_grad],
        "name": "default",
        "weight_decay": weight_decay,
    }
    return split_params_into_different_groups_for_optimizer(parameters)



from typing import List, Union

from torch import nn

from internlm.core.context import ParallelMode
from internlm.core.context import global_context as gpc
from internlm.core.parallel.shard import pipeline_parallel_sharding_wrapper
from internlm.model.registry import model_initializer
from internlm.utils.common import get_current_device


def create_model(model_type, *args, **kwargs) -> Union[nn.Module, List[nn.Module]]:
    num_layers = kwargs.pop("num_layers")
    num_chunks = kwargs.pop("num_chunks", 1)

    # TODO: fix use_flash_attn parameter config
    kwargs.pop("use_flash_attn", False)
    kwargs.pop("apply_post_layer_norm")
    kwargs.pop("embed_split_hidden", True)

    kwargs["checkpoint"] = float(kwargs.get("checkpoint", False))
    kwargs["device"] = get_current_device()

    model_buidler = model_initializer.get_module(module_name=model_type)

    if not gpc.is_using_parallel_mode(ParallelMode.PIPELINE):
        kwargs["first"] = kwargs["last"] = True
        kwargs["start_layer_idx"] = 0
        kwargs["num_layers"] = num_layers
        model = model_buidler(*args, **kwargs).to(kwargs["device"])
        setattr(model, "first_layer", 0)
        setattr(model, "last_layer", num_layers)
    else:
        model = pipeline_parallel_sharding_wrapper(num_layers, num_chunks, model_buidler, *args, **kwargs)

    return model



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import numpy as np
import torch
from torch.utils.data import Dataset


class RandomDatasetMultimodal(Dataset):
    """
    RandomDataset for generating random dataset.

    Args:
        num_samples (int): The number of samples to generate.
        max_len (int): The maximum length of each sample.
        image_token_id (int): The placeholder of image.
        image_size (int): The image size.
        image_patch_size (int): The patch size of vit.
        image_token_size (int): The number of placeholder of each image.

    """

    def __init__(
        self,
        num_samples=10000,
        max_len=2048,
        image_token_id=200000,
        image_size=336,
        image_token_size=(336 // 14) ** 2,
    ) -> None:
        super().__init__()
        rng = np.random.RandomState(1999)
        max_num = rng.randint(1, 30, size=(num_samples,))
        rep_num = rng.randint(10, 200, size=(num_samples,))
        data = []
        lengths = []
        images = [
            [torch.randn((3, image_size, image_size))] for _ in range(num_samples)
        ]  # num_samples x img_num x tensor(C x H x W)
        for n, r in zip(max_num, rep_num):
            d = list(range(n)) * r
            d = [n, r] + [image_token_id] * image_token_size + d
            d = d[:max_len]
            data.append(d)
            lengths.append(len(d))
        self.data = data
        self.images = images
        self.max_len = max_len
        self.lengths = np.array(lengths, dtype=int)

    def __getitem__(self, index):
        d = self.data[index]
        input_ids = np.array(d, dtype=int)
        images = self.images[index]
        return {"tokens": list(input_ids), "images": images, "type_id": 0}

    def get_dataset_name(self):
        return "dummy_path/dummy_lang/dummy_ds/train.bin"

    def __len__(self):
        return len(self.data)



from typing import Optional, Tuple

import numpy as np
import torch

from internlm.core.context import ParallelMode
from internlm.core.context import global_context as gpc
from internlm.model.moe.base_layer import BaseMoELayer
from internlm.model.moe.megablock.megablock_moe import MegaBlockMoE
from internlm.model.moe.megablock.mlp import MegaBlockGroupedFeedForward
from internlm.model.moe.megablock.utils import promote_scalar

try:
    import stk
    from megablocks import ops
except (ModuleNotFoundError, ImportError):
    stk = None
    megablocks = None


class MegaBlockdMoE(MegaBlockMoE):
    """
    Built on the paper and library Megablocks as described in
    https://arxiv.org/abs/2211.15841. This implementation is
    strictly equivalent to standard MoE with full capacity (no
    dropped tokens). It's faster since it formulates MoE operations
    in terms of block-sparse operations to accomodate imbalanced
    assignments of tokens to experts, whereas standard MoE either
    (1) drop tokens at the cost of reduced performance or (2) set
    capacity factor to number of experts and thus waste computation
    and memory on padding.
    """

    def __init__(  # pylint: disable=W0231
        self,
        hidden_size: int,
        ep_group: Optional[torch.distributed.ProcessGroup],
        ep_size: int,
        num_experts: int,
        top_k: int = 1,
        parallel_mode: str = "tensor",
        device: Optional[torch.device] = None,
        dtype: Optional[torch.device] = None,
        multiple_of: int = 256,
    ) -> None:
        assert gpc.expert_parallel_size == 1, "do not support expert parallel"
        self.top_k = top_k
        self.num_experts = num_experts

        tp_size = gpc.get_world_size(ParallelMode.TENSOR)
        self.ffn_dim = multiple_of * ((int(hidden_size * gpc.config.model.mlp_ratio) + multiple_of - 1) // multiple_of)
        assert self.ffn_dim % tp_size == 0
        if parallel_mode == "tensor":
            self.ffn_dim_per_row = self.ffn_dim // tp_size // ep_size
        else:
            self.ffn_dim_per_row = self.ffn_dim // ep_size
        BaseMoELayer.__init__(  # pylint: disable=W0233
            self,
            torch.nn.Linear(hidden_size, num_experts, bias=False),
            MegaBlockGroupedFeedForward(
                hidden_size,
                (self.ffn_dim // tp_size) * (num_experts // ep_size),
                parallel_mode,
                device,
                dtype,
            ),
            ep_group,
            ep_size,
            1,
        )

        # Calculate the number of bits needed to represent the expert indices
        # so that we can pass it to radix sort.
        self.sort_end_bit = max(int(np.ceil(np.log2(self.num_experts))), 1)
        self.blocking = 128
        self.quantize_scatter_num_bits = -1

        # Calculate the number of bits needed to represent the column indices
        # in the intermediate sparse matrix.
        max_column_index = (self.ffn_dim * (self.num_experts // ep_size)) // self.blocking
        self.transpose_sort_end_bit = max(int(np.ceil(np.log2(max_column_index))), 1)

        # re-init the number of experts in each device
        self.num_local_experts = num_experts // ep_size

        self.forward_fn = self._forward

    def sparse_transpose(
        self, size: int, row_indices, column_indices
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        block_columns = size[1] // self.blocking

        # Sort row indices by column indices to get the transposed matrix's
        # column indices.
        #
        # NOTE: Our sort operation uses the same width indices as the input
        # values. To avoid overflow when we have large activation matrices
        # we cast to 32-bit before sorting.
        _, gather_indices = ops.sort(column_indices.int(), self.transpose_sort_end_bit)

        # There are a constant number of blocks in every row of the sparse
        # matrix. A blocks offset is:
        #
        # row_index * blocks_per_row + column_index % blocks_per_row
        #
        # Once we have the block offsets ordered for transposition we can
        # divide by blocks_per_row to get the transposed column indices.
        column_indices_t = row_indices.gather(0, gather_indices.long())
        block_offsets_t = gather_indices.int()

        zero = torch.zeros((1,), dtype=torch.int32, device=row_indices.device)
        nnz_per_column = ops.histogram(column_indices, block_columns)
        nnz_per_column = ops.inclusive_cumsum(nnz_per_column, 0)
        offsets_t = torch.cat([zero, nnz_per_column])
        return column_indices_t, offsets_t, block_offsets_t

    def topology(self, x: torch.Tensor, padded_bins: torch.Tensor):
        padded_tokens, _ = x.size()
        assert padded_tokens % self.blocking == 0
        assert self.ffn_dim_per_row % self.blocking == 0

        # Offsets for the sparse matrix. All rows have the
        # same number of nonzero blocks dictated by the
        # dimensionality of a single expert.
        block_rows = padded_tokens // self.blocking
        blocks_per_row = self.ffn_dim_per_row // self.blocking
        offsets = torch.arange(
            0,
            block_rows * blocks_per_row + 1,
            blocks_per_row,
            dtype=torch.int32,
            device=x.device,
        )

        # Indices for the sparse matrix. The indices for
        # the intermediate matrix are dynamic depending
        # on the mapping of tokens to experts.
        column_indices = ops.topology(padded_bins, self.blocking, block_rows, blocks_per_row)

        # TODO(tgale): This is unused. Remove the need for this in stk.
        # For now, use meta init to save the device memory.
        data = torch.empty(
            column_indices.numel(),
            self.blocking,
            self.blocking,
            dtype=x.dtype,
            device="meta",
        )
        shape = (padded_tokens, self.ffn_dim_per_row * self.num_experts)
        row_indices = stk.ops.row_indices(shape, data, offsets, column_indices)
        column_indices_t, offsets_t, block_offsets_t = self.sparse_transpose(shape, row_indices, column_indices)
        return stk.Matrix(
            shape,
            data,
            row_indices,
            column_indices,
            offsets,
            column_indices_t,
            offsets_t,
            block_offsets_t,
        )

    def indices_and_padded_bins(
        self, selected_experts: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        # Sort the expert ids to produce the scatter/gather
        # indices for the permutation.
        selected_experts = selected_experts.int()
        bin_ids, indices = ops.sort(selected_experts, self.sort_end_bit)

        # Histogram the expert ids to identify the number of
        # tokens routed to each expert.
        tokens_per_expert = ops.histogram(selected_experts, self.num_experts)

        # Round the token counts up to the block size used in
        # the matrix muliplications. Caculate the starting
        # position of each bin.
        padded_tokens_per_expert = ops.round_up(tokens_per_expert, self.blocking)
        padded_bins = ops.inclusive_cumsum(padded_tokens_per_expert, 0)
        padded_bins = promote_scalar(padded_bins)

        # Calculate the bin bounds for the sorted tokens.
        bins = ops.inclusive_cumsum(tokens_per_expert, 0)
        bins = promote_scalar(bins)
        return indices, bin_ids, bins, padded_bins, tokens_per_expert

    def _forward(self, x, expert_weights, top_experts) -> torch.Tensor:
        with torch.no_grad():
            (indices, bin_ids, bins, padded_bins, tokens_per_expert) = self.indices_and_padded_bins(top_experts)

        # Permute tokens and pad to prepare expert computation
        # (top_k * sequence_length + padding, model_dim)
        x = ops.padded_gather(x, indices, bin_ids, bins, padded_bins, self.top_k)

        # Create the sparse matrix topology
        with torch.no_grad():
            topo = self.topology(x, padded_bins)

        # Perform the expert computation
        # First Dense x Dense -> Sparse for w1 and w3,
        # (top_k * sequence_length + padding, ffn_dim * n_experts)
        x = self.experts(x, topo=topo)

        # Permute back and remove padding
        # (top_k * sequence_length, model_dim)
        x = ops.padded_scatter(
            x,
            indices,
            bin_ids,
            expert_weights,
            bins,
            padded_bins,
            self.top_k,
            self.quantize_scatter_num_bits,
        )

        return x, tokens_per_expert.flatten()
